@@ -66,6 +66,11 @@ Validates OK as OpenAPI 3.0.2!
 You'll have to change the url or file reference to the location of the openapi
 document for your API.
 
+> Note: Although recursion is technically allowed under the OAS, tool support is limited
+and changing the API to not use recursion is recommended.
+At present OpenApiLibCore has limited support for parsing OpenAPI documents with
+recursion in them.
+
 If the openapi document passes this validation, the next step is trying to do a test
 run with a minimal test suite.
 The example below can be used, with `source`, `origin` and 'endpoint' altered to
@@ -123,16 +128,17 @@ import json as _json
 import sys
 from copy import deepcopy
 from dataclasses import Field, asdict, dataclass, field, make_dataclass
+from functools import cached_property
 from itertools import zip_longest
 from logging import getLogger
 from pathlib import Path
 from random import choice
-from typing import Any, Callable, Dict, List, Optional, Tuple, Type, Union
+from typing import Any, Callable, Dict, List, Optional, Set, Tuple, Type, Union
 from uuid import uuid4
 
 from openapi_core import create_spec
 from openapi_core.validation.response.validators import ResponseValidator
-from prance import ResolvingParser
+from prance import ResolvingParser, ValidationError
 from prance.util.url import ResolutionError
 from requests import Response, Session
 from requests.auth import AuthBase, HTTPBasicAuth
@@ -251,6 +257,34 @@ class RequestData:
         in_optional_headers: Callable[[str], bool] = lambda x: x in optional_headers
         return any(map(in_optional_headers, self.headers))
 
+    @cached_property
+    def headers_that_can_be_invalidated(self) -> Set[str]:
+        """
+        The header parameters that can be invalidated by violating data
+        restrictions or by not providing them in a request.
+        """
+        result = set()
+        headers = [h for h in self.parameters if h.get("in") == "header"]
+        for header in headers:
+            schema: Dict[str, Any] = header["schema"]
+            if set(schema.keys()).intersection(
+                {
+                    "enum",
+                    "minimum",
+                    "maximum",
+                    "exclusiveMinimum",
+                    "exclusiveMaximum",
+                    "minLength",
+                    "maxLength",
+                    "minItems",
+                    "maxItems",
+                }
+            ):
+                result.add(header["name"])
+            if header["required"]:
+                result.add(header["name"])
+        return result
+
     def get_required_properties_dict(self) -> Dict[str, Any]:
         """Get the json-compatible dto data containing only the required properties."""
         required_properties = self.dto_schema.get("required", [])
@@ -284,7 +318,7 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
     for an introduction.
     """
 
-    def __init__(  # pylint: disable=too-many-arguments, too-many-locals
+    def __init__(  # pylint: disable=too-many-arguments, too-many-locals, dangerous-default-value
         self,
         source: str,
         origin: str = "",
@@ -294,7 +328,10 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         password: str = "",
         security_token: str = "",
         auth: Optional[AuthBase] = None,
+        extra_headers: Optional[Dict[str, str]] = None,
         invalid_property_default_response: int = 422,
+        recursion_limit: int = 1,
+        recursion_default: Any = {},
     ) -> None:
         # region: docstring
         """
@@ -324,17 +361,46 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         A [https://docs.python-requests.org/en/latest/api/#authentication | requests AuthBase instance]
         to be used for authentication instead of the ``username`` and ``password``.
 
+        === extra_headers ===
+        A dictionary with extra / custom headers that will be send with every request.
+        This parameter can be used to send headers that are not documented in the
+        openapi document or to provide an API-key.
+
         === invalid_property_default_response ===
         The default response code for requests with a JSON body that does not comply with
         the schema. Example: a value outside the specified range or a string value for a
         property defined as integer in the schema.
+
+        === recursion_limit ===
+        The recursion depth to which to fully parse recursive references before the
+        `recursion_default` is used to end the recursion.
+
+        === recursion_default ===
+        The value that is used instead of the referenced schema when the
+        `recursion_limit` has been reached. The default `{}` represents an empty
+        object in JSON. Depending on schema definitions, this may cause schema
+        validation errors. If this is the case, `None` (`${NONE}` in Robot Framework)
+        can be tried as an alternative.
         """
         # endregion
         try:
-            parser = ResolvingParser(source, backend="openapi-spec-validator")
+
+            def recursion_limit_handler(limit, refstring, recursions):
+                return recursion_default
+
+            parser = ResolvingParser(
+                source,
+                backend="openapi-spec-validator",
+                recursion_limit=recursion_limit,
+                recursion_limit_handler=recursion_limit_handler,
+            )
         except ResolutionError as exception:
             BuiltIn().fatal_error(
-                f"Exception while trying to load openapi spec from source: {exception}"
+                f"ResolutionError while trying to load openapi spec: {exception}"
+            )
+        except ValidationError as exception:
+            BuiltIn().fatal_error(
+                f"ValidationError while trying to load openapi spec: {exception}"
             )
         if (openapi_spec := parser.specification) is None:  # pragma: no cover
             BuiltIn().fatal_error(
@@ -355,6 +421,7 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         self.auth = auth
         if username and password:
             self.auth = HTTPBasicAuth(username, password)
+        self.extra_headers = extra_headers
         self.invalid_property_default_response = invalid_property_default_response
         if mappings_path and str(mappings_path) != ".":
             mappings_path = Path(mappings_path)
@@ -659,11 +726,10 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
                 value = choice(*constrained_values)
                 if value is IGNORE:
                     continue
-                result[parameter_name] = str(value)
+                result[parameter_name] = value
                 continue
             value = value_utils.get_valid_value(parameter_schema)
-            # By the http standard, query string and header values must be strings
-            result[parameter_name] = str(value)
+            result[parameter_name] = value
         return result
 
     @keyword
@@ -773,7 +839,6 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
                 valid_url_parts.reverse()
                 invalid_url = "/".join(valid_url_parts)
                 return invalid_url
-        # TODO: add support for query parameters that can be invalidated
         raise ValueError(f"{parameterized_endpoint} could not be invalidated.")
 
     @keyword
@@ -829,14 +894,11 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
             run_keyword("ensure_in_use", url, resource_relation)
             json_data = asdict(request_data.dto)
         else:
-            try:
-                json_data = request_data.dto.get_invalidated_data(
-                    schema=request_data.dto_schema,
-                    status_code=status_code,
-                    invalid_property_default_code=self.invalid_property_default_response,
-                )
-            except ValueError as exception:
-                raise exception
+            json_data = request_data.dto.get_invalidated_data(
+                schema=request_data.dto_schema,
+                status_code=status_code,
+                invalid_property_default_code=self.invalid_property_default_response,
+            )
         return json_data
 
     @keyword
@@ -869,15 +931,17 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         headers = deepcopy(request_data.headers)
 
         if status_code == self.invalid_property_default_response:
-            parameter_names = set(params.keys()).union(set(headers.keys()))
+            parameter_names = set(params.keys()).union(
+                request_data.headers_that_can_be_invalidated
+            )
             parameter_names.update(relation_property_names)
             # if all parameters are optional and none were provided, randomly pick one
             if not parameter_names:
                 parameter_names = {
-                    d["name"]
-                    for d in request_data.parameters
-                    if d["in"] in ["query", "header"]
+                    d["name"] for d in request_data.parameters if d["in"] in ["query"]
                 }
+            if not parameter_names:
+                raise ValueError("Headers cannot be invalidated.")
         else:
             # non-default status_codes can only be the result of a Relation
             parameter_names = relation_property_names
@@ -916,6 +980,7 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
             valid_value = params[parameter_to_invalidate]
         else:
             valid_value = headers[parameter_to_invalidate]
+
         invalid_value = value_utils.get_invalid_value(
             value_schema=parameter_data["schema"],
             current_value=valid_value,
@@ -925,7 +990,7 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         if parameter_to_invalidate in params.keys():
             params[parameter_to_invalidate] = invalid_value
         else:
-            headers[parameter_to_invalidate] = str(invalid_value)
+            headers[parameter_to_invalidate] = invalid_value
         return params, headers
 
     @staticmethod
@@ -1071,11 +1136,14 @@ class OpenApiLibCore:  # pylint: disable=too-many-instance-attributes
         > Note: provided username / password or auth objects take precedence over token
             based security
         """
+        headers = headers if headers else {}
+        if self.extra_headers:
+            headers.update(self.extra_headers)
         # if both an auth object and a token are available, auth takes precedence
         if self.security_token and not self.auth:
             security_header = {"Authorization": self.security_token}
-            headers = headers if headers else {}
             headers.update(security_header)
+        headers = {k: str(v) for k, v in headers.items()}
         response = self.session.request(
             url=url,
             method=method,
